//
//   Copy on write vector class definition.
//
#ifndef CWVECT_HH
#define CWVECT_HH

#define CWVEC_STATS

#ifdef __GNUG__
#pragma interface
#endif

#ifdef CWVEC_STATS
class CWStat_struct {
public:
  CWStat_struct(void);
  ~CWStat_struct(void);
  long alloc_count;
  long delete_count;
  long copy_count;
};

extern CWStat_struct CWStats;
#endif


/**  The CWVec classes allow data to be 
  *  shared by more than one Data vector without forcing a copy. This is 
  *  especially useful for reducing overheads when passing ownership of 
  *  long data vectors. A data vector can be copied using the Clone()
  *  method, and when write access is needed, Access() is used.
  *  @memo Copy on write data vector class template.
  */
template<class T>
class CWVec {
public:
  typedef unsigned long size_type;
  typedef T    value_type;

  class vec_node {
  public:
    vec_node(size_type N, const T* dval=0, bool cpy=true);
    ~vec_node(void);
    T* ref(void) {return _data;}
    const T* ref(void) const {return _data;}
    void copy(const vec_node& v, size_type inx, size_type N);
    bool release(void) const {return !(_ref_count--);}
    size_type size(void) const {return _size;}
    bool shared(void) const {return _ref_count || !_mine;}
    void use(void) const {_ref_count++;}
  private:
    mutable size_type _ref_count;
    bool      _mine;
    size_type _size;
    T* _data;
  };

  /**  Construct a data vector of Length L. If the data pointer is non-null,
    *  the indicated data are copied into the vector. Otherwise the vector
    *  is left uninitialized.
    *  @memo Basic constructor.
    *  @param L    Length of the vector to be created in data words.
    *  @param data Optional initial data values.
    */
  explicit CWVec(size_type L=0, const T* data=0, bool cpy=true);

  /**  Constructor a new CWVec identical to the first argument, but 
    *  copy only the specified number of data words.
    *  @memo Limited copy constructor.
    */
  explicit CWVec(const CWVec& x);

  /**  Delete the vector.
    *  @memo Destructor.
    */
  ~CWVec(void);

  /**  Get a pointer to a writeable copy of the vector. This function
    *  results in copying the vector if there is more than one user
    *  of this data.
    *  @memo Request a writeable copy of the data vector.
    *  @return Pointer to a writeable CWVec with the same contents.
    */
  void access(void);

  /**  Get the capacity of the internal vector.
    *  @memo Internal vector capacity
    *  @return Vector capacity.
    */
  size_type capacity(void) const;

  /**  Clear the vector by zeroing the data length. Note that the data vector 
    *  is not released.
    *  @memo Zero the data length.
    */
  void clear(void);

  /**  The current vec_node is released and deleted if its use count falls to
    *  zero. The vec_node is address is then set to the specified address. 
    *  Note that the argument vec_node must have its instance count set 
    *  appropriately before it is passed to replace().
    *  @memo Replace the vector node.
    *  @param v pointer to the new vector node.
    */
  void replace(vec_node* v);

  /**  Release a read instance. Delete the vector if it is unused.
    *  @memo Release a read instance.
    */
  void release(void);

  /** Get the number of words allocated by the CW vector.
    * @memo Get the CW vector size.
    */
  size_type size(void) const;

  /**  Get a constant reference to the CW data.
    *  @memo Get a constant reference to the data vector.
    */
  const T* ref(void) const;

  /**  Get a writable reference to the CW data.
    *  @memo Get a writable reference to the data vector.
    */
  T* ref(void);

  /**  Insure that the specified number of data words are available in the
    *  data vector. If the requested data length is less than the current 
    *  data length, no action is taken. If the requested data length is 
    *  greater than the current length, a new vector is allocated and the 
    *  existing data are copied.
    *  @memo Insure that a vector of the specified size is available.
    */
  void reserve(size_type N);

  /**  Extend the vector to the specified length. Existing data are copied 
    *  if necessary, but added words are not initialized. If the requested 
    *  data length is less than the current length, no action is taken other
    *  than updating the current length. If the requested length is zero,
    *  the index is also set to zero. Resize does not request write access.
    *  @memo Shrink or extend the vector with uninitialized data.
    */
  void resize(size_type N);

  /**  Return true if internal data vector is shared with another CWVec.
    *  @memo Test whether vector is shared.
    */
  bool shared(void) const;

  /**  Modify CWVec limits to contain a substring of the current vector.
    *  If len is zero or inx+len is greater than the current string length, 
    *  the substring extends to the end of the current string.
    *  @memo take a substring of the current vector.
    */
  void substr(size_type inx=0, size_type len=0);

  /**  Copy a constant vector.
    *  @memo Get a constant reference to a data word.
    */
  CWVec& operator=(const CWVec& v);

  /**  Get a constant reference to a data word.
    *  @memo Get a constant reference to a data word.
    */
  const T& operator[](size_type i) const;

  /**  Get a writable reference to a data word.
    *  @memo Get a writable reference to a data word.
    */
  T& operator[](size_type i);

private:
  /**  Number of data words used.
   */
  size_type mLength;

  /**  First data word used.
   */
  size_type mOffset;

  /**  Pointer to the data vector.
   */
  vec_node*  mData;
};

//--------------------------------------  Inline implementation templates
template<class T>
inline
CWVec<T>::vec_node::vec_node(size_type N, const T* dat, bool cpy)
  : _ref_count(0), _mine(cpy), _size(N), _data(0) 
{
    if (_mine) {
        _data = new T[N];
	if (dat) memcpy(_data, dat, N*sizeof(T));
#ifdef CWVEC_STATS
	CWStats.alloc_count++;
#endif
    } else {
        _data = const_cast<T*>(dat);
    }
}

template<class T>
inline
CWVec<T>::vec_node::~vec_node(void) {
    if (_mine) {
	delete[] _data;
#ifdef CWVEC_STATS
	CWStats.delete_count++;
#endif
    }
}

template<class T>
inline void 
CWVec<T>::vec_node::copy(const vec_node& v, size_type inx, size_type N) {
    memmove(_data, v._data + inx, N*sizeof(T));
}

template<class T>
inline
CWVec<T>::CWVec(size_type L, const T* dat, bool cpy)
  : mOffset(0)
{
    mData = new vec_node(L, dat, cpy);
    mLength = L; // used to be mData->size()
}

template<class T>
inline void
CWVec<T>::substr(size_type inx, size_type len) {
    if (inx > mLength) inx = mLength;
    mOffset += inx;
    if (!len || inx+len > mLength) mLength -= inx;
    else                           mLength  = len;
}

template<class T>
inline
CWVec<T>::CWVec(const CWVec<T>& x) 
  : mData(0)
{
    *this = x;
}

template<class T>
inline void
CWVec<T>::replace(vec_node* v) {
    if (mData && mData->release()) delete mData;
    mData = v;
}

template<class T>
inline void
CWVec<T>::release(void) {
    replace(0);
}

template<class T>
inline CWVec<T>& 
CWVec<T>::operator=(const CWVec& v) {
    if (v.mData) v.mData->use();
    replace(v.mData);
#ifdef CWVEC_STATS
    CWStats.copy_count++;
#endif
    mLength = v.mLength;
    mOffset = v.mOffset;
    return *this;
}

template<class T>
inline
CWVec<T>::~CWVec() {
    replace(0);
}

template<class T>
inline void
CWVec<T>::access(void) {
    if (mData->shared()) {
        replace(new vec_node(mLength, mData->ref()+mOffset));
	mOffset = 0;
    }
}

template<class T>
inline void
CWVec<T>::clear(void) {
    mLength = 0;
    mOffset = 0;
    if (shared()) replace(0);
}

template<class T>
inline typename CWVec<T>::size_type
CWVec<T>::capacity(void) const {
    if (!mData) return 0;
    return mData->size();
}

template<class T>
inline typename CWVec<T>::size_type
CWVec<T>::size(void) const {
    return mLength;
}

template<class T>
inline const T*
CWVec<T>::ref(void) const {
    return mData->ref() + mOffset;
}

template<class T>
inline T*
CWVec<T>::ref(void) {
    access();
    return mData->ref() + mOffset;
}

template<class T>
inline void
CWVec<T>::reserve(size_type N) {
    if (shared() || N > mData->size()) {
	size_type L = (N < mLength) ? N : mLength;
	vec_node* v = new vec_node(N);
	if (L) v->copy(*mData, mOffset, L);
	replace(v);
	mOffset = 0;
    } else if (mOffset + N > mData->size()) {
	mData->copy(*mData, mOffset, mLength);
	mOffset = 0;
    }
}

template<class T>
inline void
CWVec<T>::resize(size_type N) {
    if (!N) {
        clear();
    } else {
        reserve(N);
	mLength = N;
    }
}

template<class T>
inline bool
CWVec<T>::shared(void) const {
    return !mData || mData->shared();
}

template<class T>
inline const T& 
CWVec<T>::operator[](size_type i) const {
    return ref()[i];
}

template<class T>
inline T&
CWVec<T>::operator[](size_type i) {
    return ref()[i];
}
#endif
