#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <strings.h>
#include <iostream>
#include "web/webclient.hh"
#include "sockutil.h"
#include "Time.hh"
#include "Interval.hh"
#ifdef HAVE_ZUTIL
#include "zlib.h"
#include "zutil.h"
#endif

namespace web {
   using namespace std;


//////////////////////////////////////////////////////////////////////////
//                                                                      //
// Constants                                                            //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

   const int _CONNECT_TIMEOUT = 20; // timeout for connecting to web server
   const int _TIMEOUT = 30; // timeout for web access

   const int gz_magic[2] = {0x1f, 0x8b}; /* gzip magic header */
   const char* const html_get = "GET %s HTTP/1.0\r\n"
   "Accept-Encoding: gzip\r\n\r\n";
   // const char* const html_get = "GET %s HTTP/1.0\r\n\r\n";


//////////////////////////////////////////////////////////////////////////
//                                                                      //
// Utility functions                                                    //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   static string trim (const char* p)
   {
      while (isspace (*p)) p++;
      string s = p;
      while ((s.size() > 0) && isspace (s[s.size()-1])) {
         s.erase (s.size() - 1);
      }
      return s;
   }

//______________________________________________________________________________
   string http_request::mangle (const string& s)
   {
      const char* mark = "-_.!~*'()";
      const char* hex = "0123456789ABCDEF";
      string::size_type pos = 0;
      string r = s;
      while (pos < r.size()) {
         char c = r[pos];
         if (!isalnum (c) && (strchr (mark, c) == 0)) {
            string encode = 
               "%" + string (1, hex[c >> 4]) + string (1, hex[c&0x0F]);
            r.erase (pos, 1);
            r.insert (pos, encode);
         }
         ++pos;
      }
      return r;
   }

//____________________________________________________________________________
  static int getxdig(char d) {
    int r=-1;
    if (d <= '9' && d >= '0') {
      r = d - '0';
    } else if (d <= 'f' && d >= 'a') {
      r = d - 'a' + 10;
    } else if (d <= 'F' && d >= 'A') {
      r = d - 'A' + 10;
    }
    return r;
  }

   string http_request::demangle (const string& s)
   {
     string r = s;
     string::size_type j = 0;
     string::size_type n = r.size();
     for (string::size_type i=0; i < n; i++) {
       char decode = r[i];
       if (decode == '%' && i+2 < n) {
	 int x1 = getxdig(r[i+1]);
	 if (x1 >= 0) {
	    int x2 = getxdig(r[i+2]);
	    if (x2 >= 0) {
	        decode = char(x1*16 + x2);
		i += 2;
	    }
	 }
       }
       r[j++] = decode;
     }
     if (j<n) r.erase(j, n-j);
     return r;
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// http_request                                                         //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   bool http_request::header_sort::operator () (const std::string& s1, 
                     const std::string& s2) const 
   {
      return strcasecmp (s1.c_str(), s2.c_str()) < 0; 
   }

//______________________________________________________________________________
   http_request::http_request()
   : fData (0), fLen (0), fSock (-1)
   {
   }

//______________________________________________________________________________
   http_request::http_request (const char* url, const char* addr, int port)
   : fData (0), fLen (0), fSock (-1)
   {
      request (url, addr, port);
   }

//______________________________________________________________________________
   http_request::~http_request()
   
   {
      if (fData) delete [] fData;
      if (fSock >= 0) ::close (fSock);
   }

//______________________________________________________________________________
   bool http_request::open (const char* addr, int port)
   {
      // connect to the name server
      fSock = ::socket (PF_INET, SOCK_STREAM, 0);
      if (fSock == -1) {
         return false;
      }
      struct sockaddr_in name;		/* fSocket name */
      name.sin_family = AF_INET;
      name.sin_port = htons (port);
      if (nslookup (addr, &name.sin_addr) < 0) {
         ::close (fSock);
         fSock = -1;
         return false;
      }
         // connect to the web server
      timeval timeout = {_CONNECT_TIMEOUT, 0};
      if (connectWithTimeout (fSock, (struct sockaddr*) &name, 
                           sizeof (name), &timeout) < 0) {
         ::close (fSock);
         fSock = -1;
         return false;
      }
      return true;
   }

//______________________________________________________________________________
   void http_request::close ()
   {
      if (fSock >= 0) {
         ::close (fSock);
         fSock = -1;
      }
   }

//______________________________________________________________________________
   bool http_request::sendRequest (const char* url)
   {
       // send request
      char cmd[16*1024];
      sprintf (cmd, html_get, url);
      // cout << "***********************" << endl;
      // cout << cmd;
      // cout <<  "***********************" << endl;
      int status = ::send (fSock, cmd, strlen (cmd), 0);
      if (status == -1) { 
         ::close (fSock);
         fSock = -1;
         return false;
      }
      return true;
   }

//______________________________________________________________________________
   bool http_request::recvHeader ()
   {
      const int bufsize = 1024;
      bool emptyline = false;
      bool firstline = true;
      char buf[bufsize+10];
      memset (buf, 0, sizeof (buf));
      int len = 0;
      int start = 0;
   
      Time t0 = Now();
      do {
         // check timeout
         Time t1 = Now();
         if (t1 - t0 > Interval (_TIMEOUT)) {
            cerr << "Timeout during read" << endl;	
            return false;
         }
         // set descriptor to non-blocking
         int flags;
         if ((flags = fcntl (fSock, F_GETFL, 0)) == -1) {
            cerr << "Unable to get fd flags" << endl;
            return false;
         }
         if (fcntl (fSock, F_SETFL, flags | O_NONBLOCK) == -1) {
            cerr << "Unable to set fd flags" << endl;
            return false;
         }
      	 // shift data if necessary
         if (start > 0) {
            memmove (buf, buf + start, len);
            start = 0;
         }
         // read as many characters as possible
         int n = ::read (fSock, buf + len, bufsize - len);
         // set descriptor back to its original state
         if (fcntl (fSock, F_SETFL, flags) == -1) {
            cerr << "Unable to set fd flags" << endl;
            return false;
         }
         // continue if nothing was read
         if ((n == 0) || ((n == -1) && (errno == EAGAIN))) {
            const timespec tick = {0, 500000000};
            nanosleep (&tick, 0);
            continue;
         }
         if (n < 0) {
            cerr << "Read error " << n << " " << errno << endl;
            return false;
         }
         len += n;
         buf[len] = 0;
         // cout << endl << "HEADER (" << len << ")" << endl;
         // cout << "++++++++++++++++++++++++++++++++++" << endl;
         // cout.write (buf, len);
         // cout << endl << "++++++++++++++++++++++++++++++++++" << endl;
         // analyze return string
         do {
            string line;
            char* p = strstr (buf + start, "\r\n");
            if (p) {
               *p =0;
               line = trim (buf + start);
               if (!firstline) emptyline = line.empty();
               int diff = p - (buf + start) + 2;
               start += diff;
               len -= diff;
            } 
            else {
               break;
            }
            if (!line.empty()) {
               // Status line
               if (firstline) {
                  // parse request
                  if (line.find ("200 OK") == string::npos) {
                     return false;
                  }
               }
               // header information
               else {
                  string::size_type pos = line.find (":");
                  if (pos != string::npos) {
                     string name (line.begin(), line.begin() + pos);
                     line.erase (0, pos + 1);
                     string token = trim (line.c_str());
                     if (!name.empty()) {
                        fHeader[name] = token;
                     }
                  }
               }
               firstline = false;
            }
         } while (!emptyline);
      } while (!emptyline);
   
      // set data length and allocate data array
      header_type::iterator contlen = fHeader.find ("Content-length");
      fLen = 0;
      if (contlen != fHeader.end()) {
         fLen = atoi (contlen->second.c_str());
         if (fLen < 0) fLen = 0;
      }
      fData = new char[fLen+10];
      if (len > 0) {
         if (len > fLen) len = fLen;
         memcpy (fData, buf + start, len);
         fData[len] = 0;
      }
      fFill = len;
      return true;
   }

//______________________________________________________________________________
   bool http_request::recvContent ()
   {
      Time t0 = Now();
      int len = fFill;
      while (len < fLen) {
         // check timeout
         Time t1 = Now();
         if (t1 - t0 > Interval (_TIMEOUT)) {
            cerr << "Timeout during read" << endl;	
            return false;
         }
         // set descriptor to non-blocking
         int flags;
         if ((flags = fcntl (fSock, F_GETFL, 0)) == -1) {
            cerr << "Unable to get fd flags" << endl;
            return false;
         }
         if (fcntl (fSock, F_SETFL, flags | O_NONBLOCK) == -1) {
            cerr << "Unable to set fd flags" << endl;
            return false;
         }
         // read as many characters as possible
         int n = ::read (fSock, fData + len, fLen - len);
         // set descriptor back to its original state
         if (fcntl (fSock, F_SETFL, flags) == -1) {
            cerr << "Unable to set fd flags" << endl;
            return false;
         }
         // continue if nothing was read
         if ((n == 0) || ((n == -1) && (errno == EAGAIN))) {
            const timespec tick = {0, 500000000};
            nanosleep (&tick, 0);
            continue;
         }
         if (n < 0) {
            cerr << "Read error " << n << " " << errno << endl;
            return false;
         }
         len += n;
         fFill += n;
      };
   
      fData[fLen] = 0; // make sure we terminate with 0
      return true;
   }

#ifdef HAVE_ZUTIL
//////////////////////////////////////////////////////////////////////////
//                                                                      //
// uncompress_gzip  (similar to uncompress but read a a gzip 'file'     //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   int ZEXPORT uncompress_gzip (Bytef *dest, uLongf *destLen,
                     const Bytef *source, uLong sourceLen)
   {
      z_stream stream;
      int err;
   
      // check header
      if ((source[0] != gz_magic[0]) || (source[1] != gz_magic[1])) {
         return Z_DATA_ERROR;
      }
   
      // decode length
      uLong len = 0;
      uLong end = sourceLen - 1L;
      for (int n = 0; n < 4; ++n, --end) {
         len <<= 8;
         len |= (int)(source[end] & 0xff);
      }
      if (!dest) {
         *destLen = len;
         return Z_OK;
      }
      // make sure we have enough space
      if (len > *destLen) {
         return Z_BUF_ERROR;
      }
   
      // inflate
      stream.next_in = (Bytef*)source + 10L;
      stream.avail_in = (uInt)sourceLen - 18L;
      stream.next_out = dest;
      stream.avail_out = (uInt)*destLen;
      if ((uLong)stream.avail_out != *destLen) 
         return Z_BUF_ERROR;
   
      stream.zalloc = (alloc_func)0;
      stream.zfree = (free_func)0;
      stream.opaque = (voidpf)0;
   
      err = ::inflateInit2(&stream, -MAX_WBITS);
      if (err != Z_OK) 
         return err;
   
      err = ::inflate(&stream, Z_FINISH);
      if (err != Z_STREAM_END) {
         ::inflateEnd(&stream);
         return err == Z_OK ? Z_BUF_ERROR : err;
      }
      *destLen = stream.total_out;
      err = ::inflateEnd(&stream);
   
      // check CRC & inflated length
      if (err == Z_OK) {
         uLong crc = 0;
         uLong end = sourceLen - 5L;
         for (int n = 0; n < 4; ++n, --end) {
            crc <<= 8;
            crc |= (int)(source[end] & 0xff);
         }
         if (crc != crc32 (0L, dest, *destLen)) {
            return Z_DATA_ERROR;
         }
         if (len != *destLen) {
            return Z_DATA_ERROR;
         }
      }
   
      return err;
   }
#endif

//______________________________________________________________________________
   bool http_request::uncompress ()
   {
#ifdef HAVE_ZUTIL
      // get length of uncompresses data
      unsigned long newlen = 0;
      if (uncompress_gzip ((unsigned char*)0, &newlen, 
                          (unsigned char*)fData, fLen) != Z_OK) {
         return false;
      }
      // inflate
      char* newdata = new char [newlen + 10];
      if (uncompress_gzip ((unsigned char*)newdata, &newlen, 
                          (unsigned char*)fData, fLen) != Z_OK) {
         delete [] newdata;
         return false;
      }
      newdata[newlen] = 0; // make sure we terminate with 0
      // cout << "Size = " << fLen << " NewSize = " << newlen << endl;
      // cout << "Dempression ratio = " << (float)newlen/fLen << endl;
      delete [] fData;
      fData = newdata;
      fLen = newlen;
      fFill = newlen;
      return true;
#else
      return false;
#endif
   }

//______________________________________________________________________________
   bool http_request::request (const char* url)
   {
      if (!url || (fSock < 0)) {
         return false;
      }
   
      // send HTTP request
      if (!sendRequest (url)) {
         return false;
      }
   
      // receive header of response
      if (!recvHeader()) {
         ::close (fSock);
         fSock = -1;
         return false;
      }
   
      // receive content
      if (!fData || !recvContent()) {
         delete [] fData;
         ::close (fSock);
         fSock = -1;
         return false;
      }
      // unzip if supported
      header_type::const_iterator encod =
         fHeader.find ("Content-Encoding");
      if ((encod != fHeader.end()) &&
         (encod->second.find ("gzip") != string::npos)) {
         if (!uncompress()) {
            return false;
         }
      }
   
      return true;
   }

//______________________________________________________________________________
   bool http_request::request (const char* url, 
                     const char* addr, int port)
   {
      if (!url || !addr) {
         return false;
      }
   
      if (fSock >= 0) {
         ::close (fSock);
         fSock = -1;
      }
      if (!open (addr, port)) {
         return false; 
      }
   
      return request (url);
   }


}
