#include "DAQSocket.hh"
#include "laudio.hh"
#include "SigFlag.hh"
#include <iostream>
#include <stdexcept>
#include <cmath>

using namespace std;

class DAQListen {
public:
  DAQListen(bool debug=false);
  ~DAQListen(void);
  void setNDS(const char* servername);
  void setAudio(const char* devicename);
  int getChannelStats(const char* Channel) throw(runtime_error);
  int getChannelRate(const char* Channel) throw(runtime_error);
  /**  Get the number of bytes in each data word.
   */
  int getChannelSize(const char* Channel) throw(runtime_error);
  void Listen(const char* Channel) throw(runtime_error);
  void setTransform(double bias, double scale);
  bool isTranSet(void) const;
private:
  DAQSocket NDS;
  SigFlag   mTerm;
  LAudio    mSpeaker;
  double    dTime;
  double    mBias;
  double    mScale;
  bool      mDebug;
  bool      mTranSet;
  bool      mChanValid;
  DAQDChannel mChanData;
};

//======================================  Channel rate accessor
inline int
DAQListen::getChannelRate(const char* channel) throw(runtime_error) {
    if (getChannelStats(channel)) return -1;
    return mChanData.mRate;
}

//======================================  Channel data size accessor
inline int
DAQListen::getChannelSize(const char* channel) throw(runtime_error) {
    if (getChannelStats(channel)) return 1;
    return mChanData.mBPS;
}

//======================================  Main program.
int 
main (int argc, const char* argv[]) {
    bool Syntax=false, debug=false;
    double bias(0.0), scale(0.0);
    const char* nds = getenv("LIGONDSIP");
    for (int i=1 ; i<argc-1 ; i++) {
        if (!strcmp(argv[i],"-NDS")) {
	    nds = argv[++i];
	} else if (!strcmp(argv[i], "-debug")) {
	    debug = true;
	} else if (!strcmp(argv[i], "-bias")) {
	    bias = strtod(argv[++i], 0);
	} else if (!strcmp(argv[i], "-scale")) {
	    scale = strtod(argv[++i], 0);
	} else {
	    cout << "Unrecognized argument " << argv[i] << endl;
	    Syntax = true;
	    break;
	}
    }

    if (argc < 2) {
        cerr << "Channel not specified" << endl;
	Syntax = true;
    }

    if (!nds) {
        Syntax = true;
	cerr << "Network data server not specified" << endl;
    }

    //----------------------------------  Bounce on error
    if (Syntax) {
        cerr << argv[0] << " Command syntax:" << endl;
	cerr << argv[0] << " [NDS <server-ip>] <channel>" << endl;
	return 1;
    }

    //----------------------------------  Create listener and do it.
    DAQListen dl(debug);
    try {
        dl.setNDS(nds);
        dl.setAudio(0);
	if (scale != 0.0)   dl.setTransform(bias, scale);
	else if (bias != 0) dl.setTransform(bias, 1.0);
	dl.Listen(argv[argc-1]);
    } catch (runtime_error& e) {
        cerr << "Fatal error encountered: " << e.what() << endl;
    }
}

DAQListen::DAQListen(bool debug) 
  : dTime(1.0), mBias(0.0), mScale(1.0), mDebug(debug), mTranSet(false),
    mChanValid(false)
{
    mTerm.add(SIGTERM);
    mTerm.add(SIGHUP);
    NDS.setDebug(debug);
}

DAQListen::~DAQListen(void) {
}

void
DAQListen::setNDS(const char* servername) {
    if (NDS.TestOpen()) NDS.close(); 
    NDS.open(servername);
}

void
DAQListen::setTransform(double bias, double scale) {
    mBias    = bias;
    mScale   = scale;
    mTranSet = true;
}

void
DAQListen::setAudio(const char* devicename) {
    if (mDebug) cout << "Opening audio device: " << devicename << endl;
    mSpeaker.open(devicename);
    mSpeaker.setPorts(LAudio::p_Speaker);
    mSpeaker.setVolume(1.0);
    mSpeaker.mute(false);
    if (mDebug) mSpeaker.dumpStatus(cout);
    if (mTranSet) mSpeaker.setInputTransform(mBias, mScale);
}

static const int maxDAQChans = 32768;

int
DAQListen::getChannelStats(const char* channel) throw(runtime_error) {
    DAQDChannel ChanList[maxDAQChans];
    if (!NDS.TestOpen()) throw runtime_error("Unable to open NDS slot");
    string chName(channel);
    int nChan = NDS.Available(ChanList, maxDAQChans);
    if (nChan > maxDAQChans) nChan = maxDAQChans;
    for (int i=0 ; i<nChan ; i++) {
        char* ptr = ChanList[i].mName;
	for (int j=1 ; j<40 ; j++) if (ptr[j] == ' ') ptr[j] = 0;
	if (chName == ChanList[i].mName) {
	    if (mDebug) std::cout << "Channel: " << ptr << " found with rate "
				  << ChanList[i].mRate << endl;
	    mChanData = ChanList[i];
	    mChanValid = true;
	    return 1;
	}
    }
    throw runtime_error("Channel not found");
    return 0;
}

void
DAQListen::Listen(const char* channel) throw(runtime_error) {
    double Rate = getChannelRate(channel);
    int nSample = int(Rate * dTime);
    int dType = getChannelSize(channel);
    if (!NDS.AddChannel(channel)) throw runtime_error("Add channel failed");
    if (NDS.RequestOnlineData()) throw runtime_error("Data Request failed");

    int AllocSz;
    switch (dType) {
    case 2:
        AllocSz = nSample*sizeof(short) + sizeof(int);
	break;
    case 4:
        AllocSz = nSample*sizeof(float) + sizeof(int);
	break;
    default:
        throw runtime_error("invalid data type");
    }
    char* Samples = new char[AllocSz];
    float* fPtr = (float*)Samples;
    short* sPtr = (short*)Samples;

    //----------------------------------  This reads in a protocol block
    int nData = NDS.GetData((char*)Samples, AllocSz);

    double minimum = 0.0;
    double maximum = 0.0;
    int    nSec(0), nEnt(0);
    double average(0.0), rms(0.0);
    while (!mTerm) {
        nData = NDS.GetData((char*) Samples, AllocSz);
	if (nData <= 0) break;
	nSec++;
	switch (dType) {
	case 2:
	    if (!nEnt) {
	        minimum = *sPtr;
		maximum = *sPtr;
	    }
	    nData = (nData - sizeof(int))/sizeof(short);
	    for (int i=0 ; i< nData ; i++) {
	        double x = sPtr[i];
		if (x < minimum) minimum = x;
		if (x > maximum) maximum = x;
		average += x;
		rms += x*x;
	    }
	    // mSpeaker.drain();
	    mSpeaker.reSample(Rate, sPtr, nData);
	    break;
	case 4:
	    if (!nEnt) {
	        minimum = *fPtr;
		maximum = *fPtr;
	    }
	    nData = (nData - sizeof(int))/sizeof(float);
	    for (int i=0 ; i< nData ; i++) {
	        double x = fPtr[i];
		if (x < minimum) minimum = x;
		if (x > maximum) maximum = x;
		average += x;
		rms += x*x;
	    }
	    // mSpeaker.drain();
	    mSpeaker.reSample(Rate, fPtr, nData);
	    break;
	default:
	    throw runtime_error("invalid data type");
	}
	nEnt += nData;
    }

    //----------------------------------  Print statistics.
    cout << "Listen ended after " << nSec << " Seconds" << endl;
    if (nEnt != 0) {
        cout << "    Minimum value:  " << minimum << endl;
        cout << "    Maximum value:  " << maximum << endl;
	average /= double(nEnt);
        cout << "    Average value:  " << average << endl;
	rms = rms/nEnt - average*average;
	if (rms > 0) rms = sqrt(rms);
	else         rms = 0.0;
	cout << "    RMS value:      " << rms << endl;
    }
    delete[] Samples;
}




