#ifndef TransientLeague_HH
#define TransientLeague_HH

// Class for keeping records about the biggest transient
// 1) memory of a timeseries of 1 second around the n events with
//    the biggest slope
// 3) history buffers that allow the slope detector to run continuously
//
// Edward Daw, 29th June 2001

// root includes for float histogram class. Root includes guessed first
// by looking in dmt source distribution at
// .../src/root/GUI/EvalPanel/EvalPanel.cc
// #include "TH1.h"
#include "Histogram1.hh"

// slope search includes
#include "Event.hh"
#include "MonServer.hh"

// dmt includes for TSeries classes
#include "TSeries.hh"

// key for histograms maintained by the monitor
#define ALL_HOURLY_FILTERED     1
#define ALL_HOURLY_RAW          2
#define ALL_CUMULATIVE_FILTERED 3
#define ALL_CUMULATIVE_RAW      4

/** This class stores information about transient events in the
 * interferometer. Each time a second of data is acquired it is
 * @author Edward J. Daw
 * @email edaw@lsuligo.phys.lsu.edu
 * @version 1.0; last modified July 10th 2001
 */

class TransientLeague
{
public:
/** constructor initializes timeseries buffers for biggest transients,
  * history buffers for information about previosly acquired timeseries,
  * accumulated histograms of all timeseries data
  * @param tsinit Time series used to estimate data mean, standard deviation.
  * @param nkeep Number of transients to keep in records.
  * @param keep_width_bins Number of bins considered as one transient. 
  * @param event_store_width_bins Number of bins around the maximum of a peak
  * stored in the event timeseries.
  */
  TransientLeague(TSeries tsinit, int nkeep, int keep_width_bins,
		  int event_store_width_bins, std::string monserver_name );

/** constructor initializes timeseries buffers for the biggest transients.
  * Records and history buffers for both decimated and undecimated data are
  * created by this constructor
  */
  TransientLeague(TSeries tsinit, TSeries tsraw, int nkeep, 
		  int keep_width_bins, int event_store_width_bins, 
		  std::string monserver_name );

/** Default constructor.
  */
  TransientLeague(void);

///** Copy constructor
//*/
//TransientLeague(const TransientLeague&);

/** Function to set fields to initial values in a transient league
  */
  int initialize(TSeries tsinit, int nkeep, int keep_width_bins,
		  int event_store_width_bins, std::string monserver_name);

/** Set transient league fields to initial values, including a space
  * for records of the raw (unfiltered) data 
  */
  int initialize(TSeries tsinit, TSeries tsraw, int nkeep, 
		 int keep_width_bins, int event_store_width_bins, 
		 std::string monserver_name);

/** Add data to transient database.
  */
  int search(TSeries ts, int bins_to_skip, bool cont_data_flag);

/** Add data to transient database including raw timeseries data
  */
  int search(TSeries ts, TSeries rts, int bins_to_skip, bool cont_data_flag);

/** Dump event amplitudes to standard output.
  */
  int dumpAmplitudes(void);

/** Get length of history buffer from last data
  */
  int getHistoryLength(void);

/** Dump text summary of transients to the standard output
  */
  int dumpLeague(void);

/** Dump text summary of histogram data
  */
  int outputHistogramContents(std::ostream&, Histogram1* hist);

/** Dump text summary of transients to an ostream
  */
  int outputLeague(std::ostream&, bool eseg_write);

/** Return the number of trigger events currently stored
  */
  int getNEvents(void);

/** Return time of event
  */
  Time getEventTime(int event_index);

/** Return the history timeseries surrounding an event
 */
  TSeries* getEventHistory(int event_index);

/** Return duration of event
  */
  Interval getEventDuration(int event_index);

/** Tells external function if the league has been initialized
  */
  bool inUse(void);

/** Tells external function if raw timeseries data is stored
  */
  bool rawDataStored(void);

/** Resets allocated memory in transient league
  */
  void Reset(void);

/** Emptys event list and resets allocated memory but keeps history buffer
  */
  void RestartEventList(Time start_time);

/** Return start time of league
  */
  Time startTime(void);
  
/** Add data to histgrams
  */
  int addHistData(unsigned int histident, double* data, int ndata);

/** Get pointer to histogram
  */
  Histogram1* getHistPointer(unsigned int histident);

  // default destructor
  ~TransientLeague(void);

/** Function to allow retrieval of pointers to loudest events
  */
  TSeries* monitorUpdate(int i);
  
/** Function to allow retrieval of pointers to raw timeseries
  */
  TSeries* monitorRawUpdate(int i);

/** Function to add an element to the event vector
  */
  void addEvent(Event e);

private:
  // logical for whether memory allocated to timeseries keeping
  // buffers
  bool _in_use;
  bool _history_allocated;
  bool _history_append_fail;

  // logical that is true when memory has been allocated to 
  // timeseries for raw (unfiltered) data
  bool _raw_timeseries_allocated;

  // counter for number of pushes to monitor server
  unsigned int _push_counter;

  // memory for timeseries data surrounding biggest events
  TSeries* _history_buffer;
  TSeries* _raw_history_buffer;

  // static array of pointers to time series for referencing
  // by DMT viewer
  TSeries* _current_raw_tseries;
  TSeries** _p_biggest_event_displays;
  TSeries** _p_raw_event_displays;

  // event holder
  std::vector<Event> peak_vector;

  // maximum number of events in holder
  int _max_number_of_peaks_kept;

  // width of peak in bins
  int _peak_width;

  // width of timeslice stored in event for display on viewer
  int _viewer_width;

  // start time for current league
  Time _league_start_time;

  // an initial event for use as a seed for a new league
  Event _initial_event;

  // time series storage for initial event
  TSeries* _initial_event_tseries;
  TSeries* _initial_raw_event_tseries;

  // the name appearing in the DMT viewer summary window
  std::string _monserver_name;

  // a pointer to a one dimensional histogram of filtered
  // data
  Histogram1* _filthist;
  Histogram1* _periodfilthist;
};

#endif // define TransientLeague_HH





