#include "Tunnel.hh"
#include "lmsg/Buffer.hh"
#include "lmsg/ErrorList.hh"
#include <fstream>
#include "PConfig.h"
#ifdef __GNU_STDC_OLD
#include <gnusstream.h>
#else
#include <sstream>
#endif
#include <iostream>
#include <iomanip>
#include <signal.h>

using namespace std;
using namespace lmsg;


//======================================  Main function
int 
main(int argc, const char* argv[]) {
    Tunnel x(argc, argv);
    int rc = 1;
    if (x.isActive()) rc = x.MainLoop();
    return rc;
}

//======================================  Tunnel constructor/initializer
Tunnel::Tunnel(int argc, const char* argv[]) 
  : mTerm(SIGINT), mDebug(0), mActive(false), mStatTime(300.0),
    mRequests(0), mReplies(0), mAddrErrors(0), mLookupErrors(0),
    mSendErrors(0), mReplyErrors(0)
{
    mNetAddr = 0x0a000000;
    mNetMask = 0xff000000;

    //----------------------------------  Parse arguments
    for (int i=1 ; i< argc ; i++) {
        string argi = argv[i];
	if (argi == "-debug") {
	    if (i+1 >= argc || *argv[i+1] == '-') mDebug = 1;
	    else                             mDebug = strtol(argv[++i], 0, 0);
	} else if (argi == "-netaddr") {
	    mNetAddr = strtoul(argv[++i], 0, 0);
	} else if (argi == "-netmask") {
	    mNetMask = strtoul(argv[++i], 0, 0);
	} else if (argi == "-time") {
	    mStatTime = Interval(strtod(argv[++i], 0));
	} else {
	    cerr << "Unrecognized argument: " << argi << endl;
	    return;
	}
    }

    //----------------------------------  Open the transport ports
    mTunnelServer.open(lmsg::TransportMsg::o_async);
    mTunnelClient.open(lmsg::TransportMsg::o_async);

    //----------------------------------  Register the server name
    ostringstream tName;
    tName << "_Tunnel_" << hex << setw(8) << setfill('0') << mNetAddr
	  << mNetMask << ends;
    mTunnelName = tName.str();
    mNameServer.addName(mTunnelName.c_str(), mTunnelServer.getAddr(), 
			p_Tunnel);

    //----------------------------------  Ready to roll
    mTerm.add(SIGTERM);
    struct sigaction tmp;
    tmp.sa_flags   = 0;
    tmp.sa_handler = SIG_IGN;
    sigemptyset(&tmp.sa_mask);
    sigaction(SIGIO, &tmp, NULL);

    mLastTime = Now();
    if (getenv("DMTHTMLOUT")) mStatFile = string(getenv("DMTHTMLOUT")) + "/";
    mStatFile += "Tunnel-stats.txt";
    mActive = true;
}

//======================================  Tunnel destructor
Tunnel::~Tunnel(void) {
    mNameServer.remName(mTunnelName.c_str());
}

//======================================  Loop over messages
int 
Tunnel::MainLoop(void) {
    const TransportMsg* tList[2] = {&mTunnelClient, &mTunnelServer};
    while (!mTerm) {
        Buffer* buf(0);

	//------------------------------  Wait for a message on either socket.
	error_type rc = TransportMsg::waitAny(-1.0, 2, tList);
	if (rc != Continue && rc != OK && rc != TimeOut) {
	    cerr << "Tunnel: Fatal error waiting for message. rc=" 
		 << rc << endl;
	    break;
	}
        Time now(Now());

	//------------------------------  Handle any replies.
	while (mTunnelClient.waitMsg(0.0) == OK) {
	    rc = mTunnelClient.receive(&buf);
	    if (rc == BadAddress) {
	        mAddrErrors++;
		continue;
	    } else if (rc != OK) {
	        cerr << "Tunnel: Error " << rc << " receiving msg." << endl;
		continue;
	    }
	    mReplies++;
	    MsgHeader hdr = buf->getHeader();
	    ServerID  inx = hdr.getDest().getSubProcess();
	    if (mDebug) cout << "Received reply from: " << hdr.getSource()
			     << " index: " << inx << endl;
	    if (inx > mClientMap.size()) {
	        cerr << "Tunnel: Error in reply message index." << endl;
		buf->Return();
		continue;
	    } else if (! mClientMap[inx].isValid() ) {
	        cerr << "Tunnel: Duplicate reply message index." << endl;
		buf->Return();
		continue;
	    }
	    buf->getHeader().getDest() = mClientMap[inx].getReturn();
	    mClientMap[inx].Annul();
	    // rc = mTunnelServer.send(buf);
	    rc = mTunnelServer.send(buf->getHeader(), buf->getDataAddr());
	    if (rc) {
		mReplyErrors++;
		if (mDebug) cout << "Error " << rc 
				 << " occurred forwarding reply to: " 
				 << buf->getHeader().getDest() << endl;
	        // buf->Return();
	    }
	    buf->Return(); // remove this and previous when send(buf) is public
	}

	//------------------------------  Handle any forward requests
	while (mTunnelServer.waitMsg(0.0) == OK) {
	    rc = mTunnelServer.receive(&buf);
	    if (rc != OK) {
	        cerr << "Tunnel: Error receiving msg." << endl;
		continue;
	    }
	    mRequests++;
	    MsgHeader hdr = buf->getHeader();
	    ServerID inx = hdr.getDest().getSubProcess();
	    server_iter iter = mServerMap.find(inx);
	    if (iter == mServerMap.end()) {
	        MsgAddr fwdAddr;
	        rc = mNameServer.lookupID(inx, fwdAddr);
		if (rc) {
		    cout << "Tunnel: Error in process lookup" << endl;
		    mLookupErrors++;
		    buf->Return();
		    continue;
		}
		mServerMap[inx] = fwdAddr;
		iter = mServerMap.find(inx);
	    }
	    hdr.getDest()   = iter->second;
	    ServerID cIndex = addClient(hdr.getSource());
	    if (mDebug) cout << "Forwarding message from: "<< hdr.getSource()
			     << " to: " << hdr.getDest() << " index: "
			     << cIndex << endl;
	    MsgAddr reply_addr(mTunnelClient.getAddr());
	    reply_addr.setSubProcess(cIndex);
	    mTunnelClient.setReplyAddr(reply_addr);

	    buf->getHeader() = hdr;
	    // rc = mTunnelClient.send(buf);
	    rc = mTunnelClient.send(buf->getHeader(), buf->getDataAddr()); //
	    if (rc == BadAddress) {
	        mServerMap.erase(inx);
		mAddrErrors++;
	    }
	    if (rc) {
	        mSendErrors++;
	        cout << "Error " << rc << " occurred forwarding message to "
		     << hdr.getDest() << endl;
	        // buf->Return();
	    }
	    buf->Return(); // remove this and previous when send(buf) is public
	}

	//------------------------------  Print stats
	if (now >= mLastTime + mStatTime) {
	    ofstream out(mStatFile.c_str());
	    out << "Target subnet:       0x" << hex << setw(8) 
		<< setfill('0') << mNetAddr << endl;
	    out << "Target network mask: 0x" << hex << setw(8) 
		<< setfill('0') << mNetMask << endl;
	    out << "Requests:            " << mRequests << endl;
	    out << "Replies:             " << mReplies << endl;
	    out << "Address Errors:      " << mAddrErrors << endl;
	    out << "Lookup Errors:       " << mLookupErrors << endl;
	    out << "Send Errors:         " << mSendErrors << endl;
	    out << "Reply Errors:        " << mReplyErrors << endl;
	    out.close();
	    mLastTime = now;
	}
    }
    return 0;
}

Tunnel::ServerID
Tunnel::addClient(const lmsg::MsgAddr& addr) {
    unsigned long nEnt = mClientMap.size();
    for (unsigned long i=1 ; i < nEnt ; i++) {
        if (!mClientMap[i].isValid()) {
	    mClientMap[i] = ClientRec(addr);
	    return i;
	}
    }
    if (!nEnt) nEnt++;
    mClientMap.resize(nEnt + 1);
    mClientMap[nEnt] = ClientRec(addr);
    return nEnt;
}
