#ifndef CHANNAME_HH
#define CHANNAME_HH

#include <string>
#include <stdexcept>

/**  ChanName parses and constructs standard channel names of the format
  *  #Xn:<ssid>-<locale>_<name>#, where #X# is the GW detector site, #n# 
  *  is the detector or configuration number, #<ssid># is the subsystem ID,
  *  #<locale>_# is an optional location specifier, and #<name># is the 
  *  channel specifier.
  *  @author John Zweizig
  *  @version 1.0; Last modified: January 11, 2005
  */
class ChanName {
public:
  /**  Construct an empty vchannel name.
    *  @memo Default constructor.
    */
  ChanName(void);

  /**  Construct a ChanName by parsing a channel name string.
    */
  ChanName(const std::string& name, bool locale=true);

  /**  Construct a ChanName from a channel name string using an existing
    *  channel name to define default fields.
    *  @memo Construct from a string and defaults.
    */
  ChanName(const std::string& name, const ChanName& deflt);

  /**  Get the IFO name in the format "Xn".
    *  @memo Get the ifo name.
    */
  const std::string& getIFO(void) const;

  /**  Get the subsystem name string.
    *  @memo Get the sub-system name.
    */
  const std::string& getSubSystem(void) const;

  /**  Get the channel Locale name.
   */
  const std::string& getLocale(void) const;

  /**  Get the channel specific name field.
   */
  const std::string& getName(void) const;

  /**  Get the full channel name
   */
  std::string getChannel() const;

  /** Set the IFO field of the channel name.
   */
  void setIFO(const std::string& ifo) throw(std::runtime_error);

  /** Set the SubSystem field of the channel name.
   */
  void setSubSystem(const std::string& det) throw(std::runtime_error);

  /** Set the Locale field of the channel name.
   */
  void setLocale(const std::string& ifo) throw(std::runtime_error);

  /** Set the Name field of the channel name.
   */
  void setName(const std::string& ifo) throw(std::runtime_error);

  /**  Set the channel name. The first word of the name field will be assumed
    *  to be a Locale if #locale# is true. Otherwise the Locale will be assumed
    *  to be null and the entire text after the "-" will be assigned to the 
    *  Name field.
    */
  void setChannel(const std::string& ifo, bool locale=true) 
       throw(std::runtime_error);

  /**  Set the channel name, filling in any missing fields from the .
    */
  void setChannel(const std::string& ifo, const ChanName& deflt)
       throw(std::runtime_error);

private:
  std::string mIFO;
  std::string mSubSystem;
  std::string mLocale;
  std::string mName;
};

//======================================  Inline methods
inline const std::string& 
ChanName::getIFO(void) const {
    return mIFO;
}

inline const std::string& 
ChanName::getSubSystem(void) const {
    return mSubSystem;
}

inline const std::string& 
ChanName::getLocale(void) const {
    return mLocale;
}

inline const std::string& 
ChanName::getName(void) const {
    return mName;
}

#endif // !defined(CHANNAME_HH)
