#ifndef FDFilter_Base_HH
#define FDFilter_Base_HH

#include "FSpectrum.hh"
#include "FSeries.hh"

/**  The FDFilterBase class defines an API for a frequency domain filter. 
  *  Frequency domain filters operate on either FSeries or FSpectrum
  *  class objects.
  *  @author J. Zweizig
  *  @version 1.0; Last modified September 28, 2004
  */
class FDFilterBase {

public:

  /**  Default (null) destructor for the FDFilterBase class.
    *  @memo Default constructor.
    */
  FDFilterBase(void);

  /**  Virtual destructor to allow cleaning up objects of derived classes.
    *  @memo Destroy a frequency domain filter. 
    */
  virtual ~FDFilterBase(void);

  /**  This method extracts from the specified input FSpectrum that frequency
    *  range for which the filter is defined, multiplies it by the filter 
    *  function, and writes the result to the specified output FSpectrum.
    *  The input FSpectrum must be sampled with the same frequency spacing 
    *  as the filter.
    *  @memo Filter the specified FSpectrum. 
    *  @param Int_PSD Input FSpectrum to be filtered.
    *  @param Out_PSD Filtered output FSpectrum.
    */
  virtual void Apply(const FSpectrum& Int_PSD, FSpectrum& Out_PSD) = 0;

  /**  This method extracts from the specified FSereis that frequency
    *  range for which the filter is defined, multiplies it by the filter 
    *  function, and returns.
    *  @memo Filter the specified FSeries. 
    *  @param Int_DFT Input FSeries to be filtered.
    *  @param Out_DFT Filtered output FSeries.
    */
  virtual void Apply(const FSeries& Int_DFT, FSeries& Out_DFT) = 0;

  /**  This method extracts from the specified FSpectrum that frequency
    *  range for which the filter is defined, multiplies it by the filter 
    *  function, and returns the filtered FSpectrum.
    *  The input FSpectrum must be sampled with the same frequency spacing 
    *  as the filter.
    *  @memo Filter the specified FSpectrum. 
    *  @param Int_PSD Input FSpectrum to be filtered.
    */
  virtual FSpectrum Apply(const FSpectrum& Int_PSD);

  /**  This method extracts from the specified FSpectrum that frequency
    *  range for which the filter is defined, multiplies it by the filter 
    *  function, and returns.
    *  @memo Filter the specified FSeries.
    */
  virtual FSeries Apply(const FSeries& Int_DFT);

  /**  Create an identical copy of this filter.
    *  @memo Clone the filter.
    */
  virtual FDFilterBase* clone(void) const = 0;

private:
};

//======================================  Inline methods
inline
FDFilterBase::FDFilterBase(void) {
}

inline
FDFilterBase::~FDFilterBase(void) {
}

inline FSpectrum 
FDFilterBase::Apply(const FSpectrum& Int_PSD) {
    FSpectrum out;
    Apply(Int_PSD, out);
    return out;
}

inline FSeries 
FDFilterBase::Apply(const FSeries& Int_DFT) {
    FSeries out;
    Apply(Int_DFT, out);
    return out;
}

#endif  //----- FDFilter_HH
