////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                                LPEFilter.hh                                //
//                                                                            //
//               Linear predictive error filter implementation                //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#ifndef LPEFILTER_HH
#define LPEFILTER_HH

#include "FIRFilter.hh"

/**  Linear predictive error filters are FIR filters whose coefficients
  *  are optimized in a least-squares sense to predict and remove the
  *  stationary components of a signal.  It is a useful technique for
  *  removing line frequencies or a constant background from a signal and
  *  provides a simple method to rapidly identify transient events.
  *
  *  The LPEFilter class implements a linear predictive error filter based
  *  on the FIRFilter class.  In addition to the methods of FIRFilter,
  *  LPEFilter provides the following functionality.
  *
  *  A method is provided to determine the optimal filter coefficients
  *  based on training data, while ensuring that the filter history is
  *  retained.
  *
  *  Initial training is automatically performed when the filter is applied
  *  to data for the first time.  Subsequent periodic retraining is
  *  automatically performed at predefined intervals.  In addition, the
  *  user can call the training functions directly.
  *
  *  Like FIRFilter, LPEFilter operates on successive blocks of data.  To
  *  ensure that retraining does not occur across succesive data blocks,
  *  the user specified retraining period should be an integer multiple of
  *  the block duration.
  *
  *  All data are promoted to type double prior to filtering.
  *
  *  @memo Linear Predictive Error Filter implementation.
  *  @author Shourov K. Chatterji
  *  @version 1.2; Last modified: January 14, 2005
  */
class LPEFilter : public FIRFilter {

public:
   using Pipe::apply;

/**  Construct a new filter with the given filter length, retraining
  *  period, and training length.  The filter history and start times
  *  are cleared.
  *  @memo Default constructor.
  *  @param Filter length
  *  @param Retraining period
  *  @param Training length
  */
  LPEFilter(int filterLength = 0, int trainPeriod = 0, int trainLength = 0);

/**  Creates new filter with the same filter length, retraining period,
  *  training length, sample rate, and filter coefficients as the given
  *  filter.  The filter history and start times are cleared.
  *  @memo Copy constructor.
  */
  LPEFilter(const LPEFilter& lpef);

/**  Destroys the filter and releases any allocated memory.
  *  @memo Destructor.
  */
  virtual ~LPEFilter(void);

/**  Creates new filter with the same filter length, retraining period,
  *  training length, sample rate, and filter coefficients as the given
  *  filter.  The filter history and start times are cleared.
  *  @memo Assignment operator.
  */
  LPEFilter& operator=(const LPEFilter& lpef);

/**  Filters an input TSeries and places the result in the given output
  *  TSeries.  Before filtering, training is performed if the filter has
  *  never been trained or if the retrain period has elapsed.  A non-zero
  *  return value indicates a training was performed.  The filter history
  *  is also updated.
  *  @memo Filter TSeries
  *  @param TSeries input
  *  @param TSeries output
  *  @return Non-zero if retrain occurred
  */
  int apply(const TSeries& in, TSeries& out);

/**  Determine the optimal filter coefficients based on training data.
  *  The entire TSeries is used as the training data.  The training data
  *  length must be greater than the filter length.  A non-zero return value
  *  indicates that the training was unsuccessful.
  *  @memo Determine filter coefficients
  *  @param TSeries training data
  *  @return Non-zero on error
  */
  int train(const TSeries& data);

/**  Sets the length of time (in seconds) between trainings.  To ensure
  *  that training does not occur across the boundary between successive
  *  data blocks, the retraining period should be an integer multiple of
  *  the data block duration.  A value of zero prevents automatic periodic
  *  retraining of the filter.
  *  @memo Set retraining period
  *  @param Retraining period [seconds]
  */
  void setTrainPeriod(int period);

/**  Sets the amount of data (in samples) which is used to train the filter.
  *  To determine the filter coefficients, the training algorithm requires a
  *  training length greater than the filter length.
  *  @memo Set training length
  *  @param Training length [samples]
  */
  void setTrainLength(int length);

/**  Sets the filter history by applying the filter to a TSeries.
  *  @memo Set History from TSeries
  *  @param TSeries data
  */
  void setHistory(const TSeries& in);

/**  Gets the length of time (in seconds) between trainings.
  *  @memo Get retraining period
  *  @return Retraining period [seconds]
  */
  int getTrainPeriod(void) const;

/**  Gets the amount of data (in samples) which is used to train the filter.
  *  @memo Get training length
  *  @return Training length [samples]
  */
  int getTrainLength(void) const;

/**  Gets the time of the most recent filter training.  The train time is
  *  set automatically whenever the filter is trained.
  *  @memo Get time of most recent training
  *  @return Time of last training
  */
  Time getTrainTime(void) const;

/**  Gets the length in seconds of the filter's impulse response.  Determines
  *  the dead time after an impulsive event, during which the filter output
  *  is unreliable.
  *  @memo Get the filter length in seconds.
  *  @return Filter length in seconds.
  */
  Interval getTransientTime(void) const;

private:

/**  Determines the first M autocorrelation values of the sequence x, where
  *  M is the filter length.
  *  @memo Compute autocorrelation vector
  *  @param training data vector
  *  @param autocorrelation vector
  */
  void autocorr(double* x, double* r) const;

/**  Solves the symmetric Yule-Walker problem using Levinson-Durbin
  *  recursion to determine the filter coefficients.
  *  @memo Solve symmetric Yule-Walker problem
  *  @param autocorrelation vector
  *  @param filter coefficients
  */
  void levinson(double* r, double* c) const;

/**  The number of seconds between filter trainings.  A zero value
  *  indicates that the filter should not be retrained.
  *  @memo Retraining Period [seconds]
  */
  int trainPeriod;

/**  The length of data used for training the filter.
  *  @memo Training Length [samples]
  */
  int trainLength;

/**  The time of the most recent training.
  *  @memo Last training time
  */
  Time trainTime;

};

#endif // LPEFILTER_HH
